# Copyright (C) 2019 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Primitives for Tax Support
--------------------------
"""


from tendril.utils.types.unitbase import Percentage

from tendril.schema.base import NakedSchemaObject
from tendril.schema.helpers import SchemaObjectList

from .gst import GSTMixin
from .gst import DEFAULT_TAX


class TaxDefinition(NakedSchemaObject):
    def elements(self):
        e = super(TaxDefinition, self).elements()
        e.update({
            'tax':  self._p('tax',  required=True),
            'rate': self._p('rate', required=True, parser=Percentage),
        })
        return e

    @property
    def ident(self):
        return "{0} ({1})".format(self.tax, self.rate)

    def __repr__(self):
        return "<{0} {1}>".format(self.__class__.__name__, self.ident)


class TaxDefinitionList(SchemaObjectList):
    _objtype = TaxDefinition


class TaxMixin(NakedSchemaObject, GSTMixin):
    def elements(self):
        return {'tax': self._p('tax', parser=TaxDefinitionList,
                               required=False, default=DEFAULT_TAX)}

    def override_taxes(self, taxdefinition):
        self.tax = taxdefinition

    def override_tax_rate(self, rate, tax=None):
        if not isinstance(rate, Percentage):
            rate = Percentage(rate)
        if tax:
            self._get_tax(tax).rate = rate
        else:
            for t in self.tax:
                t.rate = rate

    def _get_tax(self, name):
        for tax in self.tax:
            if tax.tax == name:
                return tax

    @property
    def unit_taxes(self):
        for tax in self.tax:
            if tax.rate == 0:
                continue
            yield (tax.ident, self.effective_price * tax.rate)

    @property
    def taxes(self):
        for tax in self.tax:
            if tax.rate == 0:
                continue
            yield (tax.ident, self.extended_price * tax.rate)

    @property
    def total_price(self):
        tp = self.extended_price
        for _, tax in self.taxes:
            tp = tp + tax
        return tp

    def reset_tax_rates(self):
        pass



# Copyright (C) 2019 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Base Pricing Primitives
-----------------------
"""

from tendril.schema.base import NakedSchemaObject
from .discount import DiscountMixin
from .tax import TaxMixin


class PricingBase(NakedSchemaObject):
    def __init__(self, *args, **kwargs):
        self.qty = kwargs.pop('qty', 1)
        self.unit = kwargs.pop('unit', 1)
        super(PricingBase, self).__init__(*args, **kwargs)

    @property
    def base_price(self):
        raise NotImplementedError

    @property
    def effective_price(self):
        return self.base_price

    @property
    def iqty(self):
        return self.qty / self.unit

    @property
    def extended_price(self):
        return self.effective_price * self.iqty

    @property
    def total_price(self):
        return self.extended_price

    def reset_qty(self):
        self.qty = self.unit

    def __repr__(self):
        return "<{0} {1} {2} {3}>" \
               "".format(self.__class__.__name__,
                         self.base_price,
                         self.effective_price,
                         self.total_price)


class SimplePricingRow(DiscountMixin, TaxMixin, PricingBase):
    def __init__(self, *args, **kwargs):
        self.price = kwargs.pop('price', 0)
        self.tax = kwargs.pop('tax', None)
        self.desc = kwargs.pop('desc')
        super(SimplePricingRow, self).__init__({}, *args, **kwargs)
        self._discounts = []

    def elements(self):
        return {}

    @property
    def base_price(self):
        return self.price

    @property
    def included_addons(self):
        return {}



# Copyright (C) 2019 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Infrastructure for Composite Pricing
------------------------------------
"""

from six import iteritems
from .base import PricingBase
from .tax import TaxMixin
from .addons import AddonMixin


class PriceCollector(TaxMixin, PricingBase, AddonMixin):
    def __init__(self, *args, **kwargs):
        self.unit = 1
        self.qty = 1
        super(PriceCollector, self).__init__({}, *args, **kwargs)
        self.tax = None
        self.items = []
        self.optional_items = []
        self.addons = {}
        self._addons = []
        self._optional_addons = []

    def elements(self):
        return {}

    @property
    def base_price(self):
        return sum([i.base_price * i.iqty for i in self.items])

    @property
    def effective_price(self):
        return sum([i.effective_price * i.iqty for i in self.items])

    @property
    def extended_price(self):
        return sum([i.extended_price for i in self.items])

    @property
    def total_price(self):
        return sum([i.total_price for i in self.items])

    @property
    def grand_total(self):
        return self.extended_price + \
               sum([x[1] for x in self.taxes]) + \
               sum(x.extended_price for x in self.included_addons)

    @property
    def discounts(self):
        collector = {}

        def _get_item_discounts(litem):
            qty = litem.iqty
            for lident, ldiscount in litem.discounts:
                if lident in collector.keys():
                    collector[lident] += ldiscount * qty
                else:
                    collector[lident] = ldiscount * qty

        for item in self.items:
            _get_item_discounts(item)

        for item in self.included_addons:
            _get_item_discounts(item)

        for ident, discount in iteritems(collector):
            yield (ident, discount)

    @property
    def taxes(self):
        collector = {}

        def _get_item_taxes(litem):
            for lident, ltax in litem.taxes:
                if lident in collector.keys():
                    collector[lident] += ltax
                else:
                    collector[lident] = ltax

        for item in self.items:
            _get_item_taxes(item)

        for item in self.included_addons:
            _get_item_taxes(item)

        for ident, tax in iteritems(collector):
            yield (ident, tax)

    @property
    def included_addons(self):
        collector = {}

        def _get_item_addons(litem):
            for laddon in litem.included_addons:
                if laddon.desc in collector.keys():
                    collector[laddon.desc].add_item(laddon)
                else:
                    collector[laddon.desc] = PriceCollector()
                    collector[laddon.desc].desc = laddon.desc
                    collector[laddon.desc].add_item(laddon)

        for item in self.items:
            _get_item_addons(item)

        for _, addon in iteritems(collector):
            yield addon

        for row in super(PriceCollector, self).included_addons:
            yield row

    @property
    def optional_addons(self):
        collector = {}

        def _get_item_optional_addons(litem):
            for laddon in litem.optional_addons:
                if laddon.desc in collector.keys():
                    collector[laddon.desc].add_item(laddon)
                else:
                    collector[laddon.desc] = PriceCollector()
                    collector[laddon.desc].desc = laddon.desc
                    collector[laddon.desc].add_item(laddon)

        for item in self.items:
            _get_item_optional_addons(item)

        for _, addon in iteritems(collector):
            yield addon

        for row in super(PriceCollector, self).optional_addons:
            yield row

    def add_item(self, item, optional=False):
        if self.tax is None:
            self.tax = item.tax
        if not optional:
            self.items.append(item)
        else:
            self.optional_items.append(item)

    def __len__(self):
        return len(self.items)



# Copyright (C) 2019 Chintalagiri Shashank
#
# This file is part of Tendril.
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

"""
Primitives for Addons Support
-----------------------------
"""


from tendril.utils.types import ParseException
from tendril.utils.types.currency import CurrencyValue
from tendril.utils.types.unitbase import Percentage

from tendril.schema.base import NakedSchemaObject
from tendril.schema.helpers import SchemaObjectSet

from .tax import TaxDefinitionList
from .tax import DEFAULT_TAX
from .base import SimplePricingRow


class AddonDefinition(NakedSchemaObject):
    def elements(self):
        e = super(AddonDefinition, self).elements()
        e.update({
            'desc':      self._p('desc'),
            'price':     self._p('price',    parser=(Percentage, CurrencyValue)),
            'tax':       self._p('tax',      required=False, parser=TaxDefinitionList),
        })
        return e

    def __repr__(self):
        return "<{0} {1}>".format(self.__class__.__name__, self.desc, self.price)


class AddonDefinitionSet(SchemaObjectSet):
    _objtype = AddonDefinition


class AddonMixin(NakedSchemaObject):
    def elements(self):
        return {'addons': self._p('addons', parser=AddonDefinitionSet,
                                  required=False, default=[])}

    def include_addon(self, addon, unit=1, qty=1, price=None,
                      tax='inherit', discount=None, optional=False):
        if addon in self.addons.keys():
            desc = self.addons[addon].desc
        else:
            desc = addon

        if price:
            if not isinstance(price, (Percentage, CurrencyValue)):
                try:
                    price = Percentage(price)
                except ParseException:
                    price = CurrencyValue(price)
        else:
            price = self.addons[addon].price

        if not optional:
            self._addons.append((desc, unit, price, tax, qty, discount))
        else:
            self._optional_addons.append((desc, unit, price, tax, qty, discount))

    @property
    def included_addons(self):
        for desc, unit, price, tax, qty, discount in self._addons:
            if isinstance(price, Percentage):
                price = self.extended_price * price

            if tax:
                if tax == 'inherit':
                    tax = self.tax
                elif not isinstance(tax, TaxDefinitionList):
                    tax = TaxDefinitionList(content=tax)
            else:
                tax = TaxDefinitionList(content=DEFAULT_TAX)
            row = SimplePricingRow(
                desc=desc, unit=unit, price=price, tax=tax, qty=qty,
                vctx=self._validation_context.child('IncludeAddon')
            )
            if discount:
                row.apply_discount(*discount)
            yield row

    @property
    def optional_addons(self):
        for desc, unit, price, tax, qty, discount in self._optional_addons:
            if isinstance(price, Percentage):
                price = self.extended_price * price

            if tax:
                if tax == 'inherit':
                    tax = self.tax
                elif not isinstance(tax, TaxDefinitionList):
                    tax = TaxDefinitionList(content=tax)
            else:
                tax = TaxDefinitionList(content=DEFAULT_TAX)
            row = SimplePricingRow(
                desc=desc, unit=unit, price=price, tax=tax, qty=qty,
                vctx=self._validation_context.child('OptionalAddon')
            )
            if discount:
                row.apply_discount(*discount)
            yield row

    def get_included_addon(self, addon):
        if addon in self.addons.keys():
            addon = self.addons[addon].desc
        for row in self.included_addons:
            if row.desc == addon:
                return row

    def get_optional_addon(self, addon):
        if addon in self.addons.keys():
            addon = self.addons[addon].desc
        for row in self.optional_addons:
            if row.desc == addon:
                return row

    def reset_addons(self):
        self._addons = []
        self._optional_addons = []
